# TODO: don't auto-close <p> tags and other tags with optional end tags
package HTML::Normalize;

#use strict; comment out for use by Perl 5.002
use HTML::Filter;
use HTML::Parser;

# This package exists to allow overriding the comment method without
# hacking the TreeBuilder package.

use vars qw(@ISA);
@ISA = qw(HTML::Filter HTML::Parser);

my $normalized_lineno = 0;
my $generator = "no";

my %isFormattingTag = map { $_ => 1 } qw(FONT TT I B BIG SMALL STRIKE S
					 U Q SUB SUP
					 EM STRONG DFN CODE SAMP KBD
					 VAR CITE ABBR ACRONYM
					 PRE
					 );

# These characteristics are according to the HTML 3.2 spec at
# http://www.w3.org/TR/REC-html32.html
# and the HTML 4.0 recommendation DTD at 
# http://www.w3.org/TR/REC-html40/sgml/dtd.html and
# http://www.w3.org/TR/REC-html40/index/elements.html

# Tags that should never have end tags
my %isNonContainerTag = map { $_ => 1 } qw(ISINDEX BASE META LINK
					   HR INPUT IMG BASEFONT BR AREA
					   PARAM COL FRAME
					  );
# Tags where the end tag is optional
my %hasOptionalEndTag = map { $_ => 1 } qw(P LI TR TD TH DT DD THEAD TFOOT
					   TBODY COLGROUP OPTION
					   BODY HEAD HTML); 

# We rely on this list being disjoint from isNonContainerTag and
# hasOptionalEndTag, becuase it's used by in_preformatted_section(),
# and we don't track the containment within the aforementioned tags.
my @preformatted_tags = qw(PRE LISTING);

# When we see a tag that is in this list, we should treat everything
# as literal text until we see the corresponding end tag. Same comment
# as that for preformatted_tags applies here.
my %isLiteralSectionTag = map { $_ => 1} qw(SCRIPT XMP);

# Formatting info for each tag: tag name, with array for
# {newline before start tag, newline after start tag, 
# newline before end tag, newline after end tag, indent contents).
# Adapted from HotMetal 4.0 defaults for compatibility with it.
my %formattingInfo = ('APPLET',     [1,1,1,1,1],
		      'AREA',       [1,0,0,0,0],
		      'BASE',       [1,0,0,0,0],
		      'BLOCKQUOTE', [1,0,0,0,1],
		      'BODY',       [0,0,1,0,1],
		      'BR',         [0,1,0,0,0],
		      'CAPTION',    [1,0,0,0,0],
		      'CENTER',     [1,0,0,0,0],
		      'DD',         [1,0,0,0,1],
		      'DIR',        [1,0,0,0,0],
		      'DIV',        [1,0,0,0,0],
		      'DL',         [1,0,1,0,1],
		      'DT',         [1,0,0,0,0],
		      'EMBED',      [1,0,0,0,0],
		      'FORM',       [1,0,0,0,0],
		      'FRAME',      [1,0,0,0,0],
		      'FRAMESET',   [1,0,0,0,0],
		      'H1',         [1,0,1,0,0],
		      'H2',         [1,0,1,0,0],
		      'H3',         [1,0,1,0,0],
		      'H4',         [1,0,1,0,0],
		      'H5',         [1,0,1,0,0],
		      'H6',         [1,0,1,0,0],
		      'HEAD',       [1,0,1,1,1],
		      'HR',         [1,0,0,0,0],
		      'HTML',       [1,0,0,0,1],
		      'LH',         [1,0,0,0,0],
		      'LI',         [1,0,0,0,1],
		      'MAP',        [1,0,0,0,0],
		      'MARQUEE',    [1,0,0,0,0],
		      'MENU',       [1,0,0,0,0],
		      'META',       [1,0,0,0,0],
		      'NOFRAME',    [1,0,0,0,0],
		      'NOFRAMES',   [1,0,0,0,0],
		      'OBJECT',     [1,0,0,0,0],
		      'OL',         [1,0,0,0,1], # XXX PROBABLY WANT 0,0,0,0
		      'OPTION',     [1,0,0,0,0],
		      'P',          [1,0,0,0,1],
		      'PARAM',      [1,0,0,0,0],
		      'PRE',        [1,0,0,0,0],
		      'SCRIPT',     [1,0,0,0,0],
		      'SELECT',     [1,0,0,0,0],
		      'TABLE',      [1,0,1,0,1],
		      'TBODY',      [1,0,0,0,0],
		      'TD',         [1,0,0,0,0],
		      'TEXTAREA',   [1,0,0,0,0],
		      'TFOOT',      [1,0,0,0,0],
		      'TH',         [1,0,0,0,0],
		      'THEAD',      [1,0,0,0,0],
		      'TITLE',      [1,0,0,0,0],
		      'TR',         [1,0,1,0,1],
		      'UL',         [1,0,0,0,1] # XXX PROBABLY WANT 0,0,0,0
		     );


sub set_indent {
    my ($self, $tag, $start) = @_;

    my $formatting_info = $formattingInfo{$tag} || [0,0,0,0,0];
    my $mult = $start ? 1 : -1;
    $self->{'indent_level'} += ($$formatting_info[4] * $mult);
    $self->{'line_sep'} = "\n" . "  " x $self->{'indent_level'};
}

sub did_nl {
    my $self = shift;
    $self->{'did_nl'};
}

sub set_did_nl {
    my ($self, $val) = @_;
    $self->{'did_nl'} = $val;
}

sub pending_nl {
    my $self = shift;
    $self->{'pending_nl'};
}

sub set_pending_nl {
    my ($self, $val) = @_;
    $self->{'pending_nl'} = $val;
}

sub pre_start_whitespace {
    my ($self, $tag) = @_;

    return if $self->in_literal_section();

    my $formatting_info = $formattingInfo{$tag} || [0,0,0,0];
    my $should_nl = $$formatting_info[0];

    if ((!$self->did_nl() && $should_nl) || $self->pending_nl()) {
	$self->output($self->{'line_sep'});
    }
    $self->set_did_nl(0); # reset
    $self->set_pending_nl(0);
}
    
sub post_start_whitespace {
    my ($self, $tag) = @_;

    return if $self->in_literal_section();

    my $formatting_info = $formattingInfo{$tag} || [0,0,0,0];
    my $should_nl = $$formatting_info[1];

    # ISO8879 says newlines after start tags and before end
    # tags should be ignored, but IE4 and Netscape4 (esp. Netscape)
    # don't do this.
    # See www.w3.org/TR/REC-html40/appendix/notes.html

    if ($should_nl) {
	$self->output($self->{'line_sep'});
	$self->set_did_nl(1);
    }
}

sub pre_end_whitespace {
    my ($self, $tag) = @_;

    return if $self->in_literal_section();

    my $formatting_info = $formattingInfo{$tag} || [0,0,0,0];
    my $should_nl = $$formatting_info[2];

    if ((!$self->did_nl() && $should_nl) || $self->pending_nl()) {
	$self->output($self->{'line_sep'});
    }
    $self->set_did_nl(0); # reset
    $self->set_pending_nl(0);
}
    
sub post_end_whitespace {
    my ($self, $tag) = @_;

    return if $self->in_literal_section();

    my $formatting_info = $formattingInfo{$tag} || [0,0,0,0];
    my $should_nl = $$formatting_info[3];

    if ($should_nl) {
	$self->output($self->{'line_sep'});
	$self->set_did_nl(1);
    }
}

sub text_only_whitespace {
    my $self = shift;

    if (!$self->did_nl()) {
	$self->output($self->{'line_sep'});
    }
    $self->set_did_nl(1);
}

sub pre_text_whitespace {
    my $self = shift;

    if (!$self->did_nl()) {
	$self->output($self->{'line_sep'});
    }
}
    
sub reset_whitespace_status {
    my $self = shift;

    $self->set_did_nl(0); # reset
}
    
sub in_preformatted_section
{
    my $self = shift;

    if ($self->in_literal_section()) {
	return 1;
    }

    my $arref = $self->{'containing_elems'};
    my %tags = map { $_ => 1 } @$arref;
    foreach (@preformatted_tags) {
	return 1 if $tags{$_};
    }
    return 0;
}

sub set_literal_section
{
    my $self = shift;
    $self->{'literal_section_tag'} = shift;
}

sub unset_literal_section
{
    my $self = shift;
    $self->{'literal_section_tag'} = '';
}

sub in_literal_section
{
    my $self = shift;

    return ($self->{'literal_section_tag'});
}

sub wordcount
{
    my $self = shift;

    $self->{'wordcount'};
}

sub wordcount_reset
{
    my $self = shift;

    $self->{'wordcount'} = 0;
}

sub wordcount_incr
{
    my $self = shift;

    ++$self->{'wordcount'};
}

sub parse_file
{
    my $self = shift;
    my ($filename, $outfilename, $logfilename) = @_;

    $self->{'filename'} = $filename;

    if ($outfilename) {
	local(*F);
	open(F, ">$outfilename") || die "Can't open $outfilename for writing: $!";
	$self->{'outfile'} = *F;
    }
    if ($logfilename) {
	local(*L);
	open(L, ">>$logfilename") || die "Can't open $logfilename for writing: $!";
	my $datetime = localtime();
	print L "\n\n<LOGENTRY FILEPATH=\"$filename\" TIME=\"$datetime\">\n";

	$self->{'logfile'} = *L;
    }
    $self->SUPER::parse_file(@_);
}

sub new
{
    $normalized_lineno = 0;
    $generator = "no";
    my $class = $_[0];
#    my $self = $class->SUPER::new(); Perl 5.002 doesn't handle this
    my $self = HTML::Parser::new($class);
    bless ($self, $class);
    $self->{'containing_elems'} = [];
    $self->{'unmatched_tags'} = [];
    $self->{'indent_level'} = 0;
    $self->{'line_sep'} = "\n  ";
    $self->{'textchunk'} = '';
    $self->{'literal_section_tag'} = '';
    $self;
}

sub start  
{ 
    my($self, $tag, $attr, $attrseq, $origtext) = @_;
    # $attr is reference to a HASH, $attrseq is reference to an ARRAY

    $self->flush_text();
    $tag = uc($tag);

#    $self->set_indent($tag, 1);

    # Add this tag to the set of containing elements
    if (!$self->in_literal_section() &&
	!$isNonContainerTag{$tag} && !$hasOptionalEndTag{$tag}) {
	my $arref = $self->{'containing_elems'};
	push(@$arref, $tag);
    }

    my ($key, $val, $outtext, @outattribs);
    foreach $key (sort(keys(%$attr))) {
	$val = $$attr{$key};
	push(@outattribs, uc($key) . "=" . "\"$val\"");
    }
    $outtext = "<" . join(" ", ($tag, @outattribs)) . ">";

    $self->pre_start_whitespace($tag);
    if ($isLiteralSectionTag{$tag}) {
	$self->set_literal_section($tag);
    }
    $self->output($outtext);
    $self->post_start_whitespace($tag);

    if ($outtext =~ /^\<META CONTENT=\".*\" NAME=\"GENERATOR\"\>$/i) {
        #print "encountered generator meta-tag \"$outtext\"\n";
        my $generatortag = $outtext;
        $generatortag =~ s/^\<META CONTENT=\"//;
        $generatortag =~ s/\" NAME=\"GENERATOR\"\>$//;
        #print "generator of this file is \"$generatortag\"\n";

        if (
               ( $generatortag =~ /Microsoft FrontPage 2\.0/ ) ||
               ( $generatortag =~ /Microsoft FrontPage 3\.0/ ) ||
               ( $generatortag =~ /Microsoft FrontPage 4\.0/ ) ||
               ( $generatortag =~ /Microsoft Word 9/ ) ||
               ( $generatortag =~ /GoLive CyberStudio 3/ ) ||
               ( $generatortag =~ /GoLive CyberStudio/ ) ||
               ( $generatortag =~ /Mozilla\/4\.04 \[en\] \(X11\; I\; SunOS 5\.5\.1 sun4m\) \[Netscape\]/ )
           ) {
            $generator = "yes";
        }
    }

    if ($outtext =~ /^\<META CLEARCASE_NORMALIZE=\".*\"\>$/i) {
        #print "encountered clearcase normalization meta-tag \"$outtext\"\n";
        my $clearcase_tag = $outtext;
        $clearcase_tag =~ s/^\<META CLEARCASE_NORMALIZE=\"//;
        $clearcase_tag =~ s/\"\>$//;
        #print "value of clearcase_normalize tag is $clearcase_tag\n";
    }
}

sub end
{
    my($self, $tag, $origtext) = @_;
    my $starttag;

    $self->flush_text();

    $tag = uc($tag);
    return if !$tag;

    if (!$self->in_literal_section() && !$hasOptionalEndTag{$tag}) {
	

#    $self->set_indent($tag, 0);

	# If it's already been output because it was in the wrong order,
	# ignore it.
	my $unmatched_ref = $self->{'unmatched_tags'};
	if ($origtext ne "__SPECIAL__" && @$unmatched_ref) {
	    my @save;
	    my $found;
	    foreach(@$unmatched_ref) {
		if ($found or ($_ ne $tag)) {
		    $save[$#save + 1] = $_;
		}
		else {
		    $found = 1;
		}
	    }
	    @$unmatched_ref = @save;
	    return if $found;
	}
	
	# Pop the starting tag and any intervening non-closed tags that
	# don't have optional end tags. (Do browsers care about interleaved
	# tags that have optional end tags?)
	do {
	    my $contelems_ref = $self->{'containing_elems'};
	    $starttag = pop(@$contelems_ref);
	    if ($starttag ne $tag) {
                # gjm, don't yap about this ???
		#$self->log_output("Didn't match container tag $starttag with end tag $tag near line number $normalized_lineno.\nContaining list: " . join(' ', @$contelems_ref) . "\n\n");
		# Print the as-yet-unmatched end tags, 
		# to account for bad interleaved 
		# start/end tags (like FrontPage produces), and remember not to
		# print them when we actually see them
		unshift(@$unmatched_ref, $starttag);
		push(@$contelems_ref, $starttag); # put it back
		$self->end($starttag, "__SPECIAL__");
	    }
	} while ($starttag and $starttag ne $tag);
	#    push (@$contelems_ref, @$unmatched_ref);
	if (!$starttag) {
	    my $message = "Unmatched end tag: $tag near line number $normalized_lineno\n";
	    $message .= "Stopped processing " . $self->{'filename'} . "\n";
	    #$self->log_output ($message);
            # don't die, just let tag be passed to output
	    #die ($message);
	}
    } # end if (!$hasOptionalEndTag($tag))

    my $outtext = "</" . $tag . ">";

    $self->pre_end_whitespace($tag);
    $self->output($outtext);
    if ($tag eq $self->in_literal_section()) {
	$self->unset_literal_section();
    }
    $self->post_end_whitespace($tag);
}

sub text {
    my ($self, $text) = @_;

    $self->{'textchunk'} .= $text;
}

sub flush_text {
    my $self = shift;
    my $elem = $self->{'textchunk'};
    my $origtext = $elem;
    my $endingwhite;

    $self->{'textchunk'} = '';
    # If text is just whitespace, preserve it (intertag space) with a newline
    if ($elem =~ /^\s+$/s) {
	$self->text_only_whitespace();
	return;
    }

    # Substitute spaces for existing newlines
    if (!$self->in_preformatted_section()) {

	$elem =~ s/\n/ /gs;

	# Substitute all spaces between 
	# sentences with a single newline.
	$elem =~ s/(\s\S+(\.|;|:)(&nbsp;)?)\s+/${1}__LINESEP__/gsx;

	# Insert newline if we need to and can.
        if ($elem =~ s/^(\s+)//) {
	    $self->pre_text_whitespace();
	}
	$self->reset_whitespace_status();

	# If this text ends in space, note that we want to
	# append a newline if this is the last of a sequence of
	# text() calls or a single text() call.
	if ($elem =~ /(\s+$)/) {
	    my $end = $1;
	    if ($end !~ /\n$/) {
		$self->set_pending_nl(1);
	    }
	}
	else {
	    $self->set_pending_nl(0);
	}
	    
	# Remove extra spaces
	$elem =~ s/[ \t]+/ /gs;

#	$elem =~ s/__LINESEP__/$self->{'line_sep'}/gsx;

	# Split sentences with greater than 8 words
	my @lines = split(/__LINESEP__/, $elem);
	my @seplines;
	my $line;
        foreach $line (@lines) {
	    $self->wordcount_reset();
	    my @linewords = split(/ /, $line);
	    my @shortlinewords;
	    my $word;
	    foreach $word (@linewords) {
		$self->wordcount_incr();
		if ($self->wordcount() > 8) {
		    my $wordcount = $self->wordcount();
		    $self->wordcount_reset();
		    $seplines[$#seplines + 1] = join(' ', @shortlinewords);
		    @shortlinewords = ($word);
		}
		else {
		    $shortlinewords[$#shortlinewords + 1] = $word;
		}
	    }
	    $seplines[$#seplines + 1] = join(' ', @shortlinewords);
	}
	$elem = join($self->{'line_sep'}, @seplines);

	$self->output($elem);
    }
    else {
	$self->reset_whitespace_status();
	$self->output($elem);
    }
}

sub eof { 
    my $self = shift;
    my $fp;

    $self->flush_text();

    if (exists($self->{'outfile'})) {
	$fp = $self->{'outfile'};
	close($fp);
	delete($self->{'outfile'});
    }
    $self->{'eof'} = 1;
    # $self->SUPER::eof(); Perl 5.002 doesn't handle this
    HTML::Parser::eof($self);
    
    if ("$generator" eq "no") {
        die("NO_MATCHING_GENERATOR");
    }
}

sub output      { 
    my ($self, $text) = @_;
    #print "#### \'$text\'";
    if ($text =~ /\n/) {
        $normalized_lineno++;
    }
    my $fp;
    if (exists($self->{'outfile'})) {
	$fp = $self->{'outfile'};
	print {$fp} $text;
    }
    else {
	print $text if (!$self->{'eof'});
    }
}

sub log_output      { 
    my ($self, $text) = @_;
    my $fp = $self->{'logfile'};
    if ($fp) {
	print {$fp} $text;
    }
    else {
	print STDERR $text;
    }
}




